
/* Chrysalide - Outil d'analyse de fichiers binaires
 * client.c - connexion à un serveur Chrysalide
 *
 * Copyright (C) 2014-2017 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Chrysalide.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "client.h"


#include <assert.h>
#include <netdb.h>
#include <poll.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <openssl/err.h>
#include <openssl/ssl.h>


#include <i18n.h>


#include "keymgn.h"
#include "protocol.h"
#include "misc/rlestr.h"
#include "../../common/io.h"
#include "../../common/xdg.h"
#include "../../core/logs.h"



/* Description de client à l'écoute (instance) */
struct _GDbClient
{
    GObject parent;                         /* A laisser en premier        */

    char *author;                           /* Utilisateur représenté      */
    //char *key_file;                         /* Accès sa la clef privée     */

    const char *name;                       /* Désignation du binaire      */

    rle_string hash;                        /* Empreinte du binaire lié    */
    GList *collections;                     /* Collections d'un binaire    */

    SSL_CTX *tls_ctx;                       /* Contexte du chiffrement     */
    char *cert_file;                        /* Fichier du certificat       */
    char *key_file;                         /* Fichier de la clef privée   */

    int fd;                                 /* Canal de communication      */
    SSL *tls_fd;                            /* Même canal, mais sécurisé   */
    char *desc;                             /* Description du lien         */

    GMutex sending_lock;                    /* Concurrence des envois      */
    GThread *update;                        /* Procédure de traitement     */

};

/* Description de client à l'écoute (classe) */
struct _GDbClientClass
{
    GObjectClass parent;                    /* A laisser en premier        */

};


/* Initialise la classe des descriptions de fichier binaire. */
static void g_db_client_class_init(GDbClientClass *);

/* Initialise une description de fichier binaire. */
static void g_db_client_init(GDbClient *);

/* Procède à la libération totale de la mémoire. */
static void g_db_client_finalize(GDbClient *);

/* Démarre réellement la connexion à la base de données. */
static bool g_db_client_start_common(GDbClient *, char *);

/* Assure l'accueil des nouvelles mises à jour. */
static void *g_db_client_update(GDbClient *);



/* Indique le type défini pour une description de client à l'écoute. */
G_DEFINE_TYPE(GDbClient, g_db_client, G_TYPE_OBJECT);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des descriptions de fichier binaire.    *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_db_client_class_init(GDbClientClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */

    object = G_OBJECT_CLASS(klass);

    object->finalize = (GObjectFinalizeFunc)g_db_client_finalize;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : client = instance à initialiser.                             *
*                                                                             *
*  Description : Initialise une description de fichier binaire.               *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_db_client_init(GDbClient *client)
{
    client->tls_ctx = NULL;
    client->cert_file = NULL;
    client->key_file = NULL;

    client->fd = -1;
    client->tls_fd = NULL;
    client->desc = NULL;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : client = instance d'objet GLib à traiter.                    *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_db_client_finalize(GDbClient *client)
{
    free(client->author);

    unset_rle_string(&client->hash);

    assert(client->tls_ctx == NULL);

    if (client->cert_file != NULL)
        free(client->cert_file);

    if (client->key_file != NULL)
        free(client->key_file);

    assert(client->tls_fd == NULL);

    if (client->desc != NULL)
        free(client->desc);

    G_OBJECT_CLASS(g_db_client_parent_class)->finalize(G_OBJECT(client));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : author      = utilisateur à représenter via le client.       *
*                kfile       = clef menant à sa clef privée.                  *
*                name        = désignation humaine du binaire associé.        *
*                hash        = empreinte d'un binaire en cours d'analyse.     *
*                collections = ensemble de collections existantes.            *
*                                                                             *
*  Description : Prépare un client pour une connexion à une BD.               *
*                                                                             *
*  Retour      : Structure mise en place ou NULL en cas d'échec.              *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GDbClient *g_db_client_new(char *author, char *kfile, const char *name, const char *hash, GList *collections)
{
    GDbClient *result;                      /* Adresse à retourner         */

    result = g_object_new(G_TYPE_DB_CLIENT, NULL);

    result->author = author;
    result->key_file = kfile;

    result->name = name;

    init_static_rle_string(&result->hash, hash);
    result->collections = collections;

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : client = client pour les accès distants à manipuler.         *
*                                                                             *
*  Description : Démarre la connexion à la base de données interne.           *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_db_client_start_internal(GDbClient *client)
{
    bool status;                            /* Bilan de la connexion       */
    struct sockaddr_un addr;                /* Adresse de transmission     */
    int ret;                                /* Bilan d'un appel            */
    char *desc;                             /* Description du serveur ciblé*/

    /* Identification du serveur à contacter */

    status = build_tmp_socket("internal-server", &addr);
    if (!status) goto gdcni_error;

    /* Création d'un canal de communication */

    client->fd = socket(AF_UNIX, SOCK_STREAM, 0);
    if (client->fd == -1)
    {
        perror("socket");
        goto gdcni_error;
    }

    ret = connect(client->fd, (struct sockaddr *)&addr, sizeof(struct sockaddr_un));
    if (ret == -1)
    {
        perror("connect");
        goto gdcsi_no_listening;
    }

    asprintf(&desc, "unix://%s", addr.sun_path);

    status = g_db_client_start_common(client, desc);

    if (!status)
        goto gdcsi_no_listening;

    return true;

 gdcsi_no_listening:

    close(client->fd);

 gdcni_error:

    client->fd = -1;

    return false;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : client = client pour les accès distants à manipuler.         *
*                host   = hôte à représenter pour le service.                 *
*                port   = port de connexion pour les clients.                 *
*                                                                             *
*  Description : Démarre la connexion à la base de données distante.          *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_db_client_start_remote(GDbClient *client, const char *host, unsigned short port)
{
    struct hostent *hp;                     /* Informations sur l'hôte     */
    struct sockaddr_in addr;                /* Adresse de transmission     */
    int ret;                                /* Bilan d'un appel            */
    char *desc;                             /* Description du serveur ciblé*/
    bool status;                            /* Bilan de la connexion       */

    /* Identification du serveur à contacter */

    hp = gethostbyname(host);
    if (hp == NULL) return false;

    addr.sin_family = hp->h_addrtype;
    memcpy(&addr.sin_addr, hp->h_addr_list[0], sizeof(struct in_addr));

    addr.sin_port = htons(port);

    /* Création d'un canal de communication */

    client->fd = socket(AF_INET, SOCK_STREAM, 0);
    if (client->fd == -1)
    {
        perror("socket");
        return false;
    }

    ret = connect(client->fd, (struct sockaddr *)&addr, sizeof(struct sockaddr_in));
    if (ret == -1)
    {
        perror("connect");
        goto gdcsr_no_listening;
    }

    asprintf(&desc, "%s:%hu", host, port);

    status = g_db_client_start_common(client, desc);

    if (!status)
        goto gdcsr_no_listening;

    return true;

 gdcsr_no_listening:

    close(client->fd);
    client->fd = -1;

    return false;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : client = client pour les accès distants à manipuler.         *
*                host   = hôte à représenter pour le service.                 *
*                port   = port de connexion pour les clients.                 *
*                                                                             *
*  Description : Démarre réellement la connexion à la base de données.        *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_db_client_start_common(GDbClient *client, char *desc)
{
    const SSL_METHOD *method;               /* Mode du canal sécurisé      */
    int ret;                                /* Bilan d'un appel            */
    packed_buffer out_pbuf;                 /* Tampon d'émission           */
    bool status;                            /* Bilan d'une opération       */
    rle_string user;                        /* Nom d'utilisateur associé   */
    GChecksum *checksum;                    /* Empreinte MD5 à signer      */
    unsigned char md5_digest[16];           /* Empreinte MD5 calculée      */
    RSA *key;                               /* Clef pour la signature      */
    unsigned char sig[RSA_USED_SIZE];       /* Signature effectuée         */
    packed_buffer in_pbuf;                  /* Tampon de réception         */
    uint32_t data;                          /* Mot de données lues         */
    DBError error;                          /* Validation de la connexion  */

    client->desc = desc;

    /**
     * Mise en place d'un environnement sécurisé.
     */

    method = TLS_client_method();

    client->tls_ctx = SSL_CTX_new(method);

    if (client->tls_ctx == NULL)
    {
        LOG_ERROR_OPENSSL;
        goto quick_error;
    }

    client->tls_fd = SSL_new(client->tls_ctx);

    if (client->tls_fd == NULL)
    {
        LOG_ERROR_OPENSSL;
        goto tls_error;
    }

    SSL_set_fd(client->tls_fd, client->fd);

    ret = SSL_connect(client->tls_fd);

    if (ret != 1)
    {
        LOG_ERROR_OPENSSL;
        goto tls_error;
    }

    /**
     * On réalise l'envoi initial ; le premier paquet doit contenir :
     *    - la commande 'DBC_HELO'.
     *    - le numéro de version du client.
     *    - l'empreinte du binaire analysé.
     *    - l'identifiant de l'utilisateur effectuant des modifications.
     *    - la signature de l'empreinte MD5 de l'identifiant.
     *
     * Tout ceci est à synchroniser avec la fonction g_db_server_listener().
     */

    init_packed_buffer(&out_pbuf);

    status = extend_packed_buffer(&out_pbuf, (uint32_t []) { DBC_HELO }, sizeof(uint32_t), true);
    if (!status) goto setup_error;

    status = extend_packed_buffer(&out_pbuf, (uint32_t []) { CDB_PROTOCOL_VERSION }, sizeof(uint32_t), true);
    if (!status) goto setup_error;

    status = pack_rle_string(&client->hash, &out_pbuf);
    if (!status) goto setup_error;

    dup_into_rle_string(&user, client->author);

    status = pack_rle_string(&user, &out_pbuf);
    if (!status) goto setup_error;

    checksum = g_checksum_new(G_CHECKSUM_MD5);
    g_checksum_update(checksum, (guchar *)get_rle_string(&user), get_rle_length(&user));
    g_checksum_get_digest(checksum, (guint8 *)md5_digest, (gsize []) { sizeof(md5_digest) });
    g_checksum_free(checksum);

    key = load_rsa_key(client->key_file, true);
    if (key == NULL) goto setup_error;

    if (!sign_md5_hash(key, md5_digest, sig))
    {
        RSA_free(key);
        goto setup_error;
    }

    RSA_free(key);

    status = extend_packed_buffer(&out_pbuf, sig, RSA_USED_SIZE, false);
    if (!status) goto setup_error;

    status = ssl_send_packed_buffer(&out_pbuf, client->tls_fd);
    if (!status) goto setup_error;

    /**
     * Le serveur doit répondre pour un message type :
     *    - la commande 'DBC_WELCOME'.
     *    - un identifiant d'erreur ('DBE_NONE', 'DBE_BAD_EXCHANGE'
     *      ou 'DBE_WRONG_VERSION' ... 'DBE_LOADING_ERROR').
     */

    init_packed_buffer(&in_pbuf);

    status = ssl_recv_packed_buffer(&in_pbuf, client->tls_fd);
    if (!status) goto comm_error;

    status = extract_packed_buffer(&in_pbuf, &data, sizeof(uint32_t), true);
    if (!status) goto comm_error;

    if (data != DBC_WELCOME)
    {
        log_variadic_message(LMT_ERROR, _("The server '%s' did not welcome us!"), desc);
        goto comm_error;
    }

    status = extract_packed_buffer(&in_pbuf, &data, sizeof(uint32_t), true);
    if (!status) goto comm_error;

    error = data;

    switch (error)
    {
        case DBE_NONE:
            log_variadic_message(LMT_INFO, _("Connected to the server '%s'!"), desc);
            break;

        case DBE_WRONG_VERSION:
            log_variadic_message(LMT_ERROR, _("The server '%s' does not use our protocol version (0x%08x)..."),
                                 desc, CDB_PROTOCOL_VERSION);
            goto comm_error;
            break;

        case DBE_XML_VERSION_ERROR:
            log_variadic_message(LMT_ERROR, _("The archive from the server '%s' does not use our protocol version (0x%08x)..."),
                                 desc, CDB_PROTOCOL_VERSION);
            goto comm_error;
            break;

        case DBE_DB_LOADING_ERROR:
            log_variadic_message(LMT_ERROR, _("The server '%s' got into troubles while loading the database...."),
                                 desc);
            goto comm_error;
            break;

        default:
            log_variadic_message(LMT_ERROR, _("The server '%s' has run into an error (%u)..."),
                                 desc, error);
            goto comm_error;
            break;

    }

    client->update = g_thread_try_new("cdb_client", (GThreadFunc)g_db_client_update, client, NULL);
    if (client->update == NULL)
    {
        log_variadic_message(LMT_ERROR, _("Failed to start a listening thread for the server '%s'!"),
                             desc);
        goto comm_error;
    }

    exit_packed_buffer(&out_pbuf);
    exit_packed_buffer(&in_pbuf);

    return true;

 comm_error:

    exit_packed_buffer(&in_pbuf);

 setup_error:

    exit_packed_buffer(&out_pbuf);

    unset_rle_string(&user);

 tls_error:

    SSL_CTX_free(client->tls_ctx);
    client->tls_ctx = NULL;

 quick_error:

    close(client->fd);
    client->fd = -1;

    return false;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : client = client pour les accès distants à manipuler.         *
*                                                                             *
*  Description : Assure l'accueil des nouvelles mises à jour.                 *
*                                                                             *
*  Retour      : NULL.                                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void *g_db_client_update(GDbClient *client)
{
    struct pollfd fds;                      /* Surveillance des flux       */
    packed_buffer in_pbuf;                  /* Tampon de réception         */
    int ret;                                /* Bilan d'un appel            */
    uint32_t tmp32;                         /* Valeur sur 32 bits          */
    bool status;                            /* Bilan d'une opération       */
    uint32_t command;                       /* Commande de la requête      */
    DBError error;                          /* Bilan d'une commande passée */
    GDbCollection *collec;                  /* Collection visée au final   */
    char *msg;                              /* Message d'erreur à imprimer */

    fds.fd = client->fd;
    fds.events = POLLIN | POLLPRI;

    init_packed_buffer(&in_pbuf);

    while (1)
    {
        ret = poll(&fds, 1, -1);
        if (ret != 1) continue;

        /* Le canal est fermé, une sortie doit être demandée... */
        if (fds.revents & POLLNVAL)
            break;

        if (fds.revents & (POLLIN | POLLPRI))
        {
            reset_packed_buffer(&in_pbuf);

            status = ssl_recv_packed_buffer(&in_pbuf, client->tls_fd);
            if (!status) goto gdcu_bad_exchange;

            status = extract_packed_buffer(&in_pbuf, &command, sizeof(uint32_t), true);
            if (!status) goto gdcu_bad_exchange;

            switch (command)
            {
                case DBC_SAVE:

                    status = extract_packed_buffer(&in_pbuf, &tmp32, sizeof(uint32_t), true);
                    if (!status) goto gdcu_bad_exchange;

                    error = tmp32;

                    if (error == DBE_NONE)
                        log_variadic_message(LMT_INFO, _("Archive saved for binary '%s'"),
                                             client->name);
                    else
                        log_variadic_message(LMT_ERROR, _("Failed to save the archive for binary '%s'"),
                                             client->name);

                    break;

                case DBC_COLLECTION:

                    status = extract_packed_buffer(&in_pbuf, &tmp32, sizeof(uint32_t), true);
                    if (!status) goto gdcu_bad_exchange;

                    collec = find_collection_in_list(client->collections, tmp32);
                    if (collec == NULL) goto gdcu_bad_exchange;

                    status = g_db_collection_unpack(collec, &in_pbuf, NULL);
                    if (!status) goto gdcu_bad_exchange;

                    break;

            }

            continue;

 gdcu_bad_exchange:

            asprintf(&msg, _("Bad reception from %s"), client->desc);

            LOG_ERROR(LMT_ERROR, msg);

            free(msg);

            break;

        }

    }

    g_db_client_stop(client);

    exit_packed_buffer(&in_pbuf);

    return NULL;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : client = client pour les accès distants à manipuler.         *
*                                                                             *
*  Description : Arrête la connexion à la base de données.                    *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void g_db_client_stop(GDbClient *client)
{
    int ret;                                /* Bilan d'un appel            */

    /* Canal de communication */

    if (client->fd != -1)
    {
        /**
         * Si la fermture est forcée, le thread de traitement va terminer en erreur.
         * Donc cette fonction sera appelée deux fois. Seule la première va affecter
         * le contexte, donc on le peut pas s'assurer de la condition suivante dans
         * tous les cas.
         */

        /*assert(client->tls_ctx == NULL);*/
        return;
    }

    ret = close(client->fd);
    if (ret == -1) perror("close");

    client->fd = -1;

    g_thread_join(client->update);

    /* Environnement TLS */

    SSL_CTX_free(client->tls_ctx);
    client->tls_ctx = NULL;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : client = client pour les accès distants à manipuler.         *
*                                                                             *
*  Description : Identifie le canal de communication pour envois au serveur.  *
*                                                                             *
*  Retour      : Descripteur de flux normalement ouvert.                      *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

int g_db_client_get_fd(GDbClient *client)
{
    g_mutex_lock(&client->sending_lock);

    return client->fd;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : client = client pour les accès distants à manipuler.         *
*                                                                             *
*  Description : Marque le canal de communication comme disponible.           *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void g_db_client_put_fd(GDbClient *client)
{
    g_mutex_unlock(&client->sending_lock);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : client = client pour les accès distants à manipuler.         *
*                                                                             *
*  Description : Effectue une demande de sauvegarde de l'état courant.        *
*                                                                             *
*  Retour      : true si la commande a bien été envoyée, false sinon.         *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_db_client_save(GDbClient *client)
{
    bool result;                            /* Bilan partiel à remonter    */
    int sent;                               /* Quantité de données traitées*/

    g_db_client_get_fd(client);

    sent = SSL_write(client->tls_fd, (uint32_t []) { htobe32(DBC_SAVE) }, sizeof(uint32_t));

    result = (sent == sizeof(uint32_t));

    g_db_client_put_fd(client);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : client    = client pour les accès distants à manipuler.      *
*                timestamp = date du dernier élément à garder comme actif.    *
*                                                                             *
*  Description : Active les éléments en amont d'un horodatage donné.          *
*                                                                             *
*  Retour      : true si la commande a bien été envoyée, false sinon.         *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_db_client_set_last_active(GDbClient *client, timestamp_t timestamp)
{
    bool result;                            /* Bilan partiel à remonter    */
    packed_buffer out_pbuf;                 /* Tampon d'émission           */

    init_packed_buffer(&out_pbuf);

    g_db_client_get_fd(client);

    result = extend_packed_buffer(&out_pbuf, (uint32_t []) { DBC_SET_LAST_ACTIVE }, sizeof(uint32_t), true);

    if (result)
        result = pack_timestamp(&timestamp, &out_pbuf);

    g_db_client_put_fd(client);

    if (result)
        result = ssl_send_packed_buffer(&out_pbuf, client->tls_fd);

    exit_packed_buffer(&out_pbuf);

    return result;

}
